#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
archive_rotator.algorithms
--------------------------

Rotation algorithms are defined here.
"""

import abc
import logging


class RotatorBase(object):
    """
    Rotator interface.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def id_to_slot(self, rotation_id):
        """Given a backup id, assign a rotation slot."""
        return


class SimpleRotator(RotatorBase):
    """FIFO implementation."""

    def __init__(self, num_rotation_slots, verbose=False):
        self.num_rotation_slots = num_rotation_slots
        if verbose:
            logging.info("Rotation slots are: %s" % range(num_rotation_slots))

    def id_to_slot(self, rotation_id):
        return rotation_id % self.num_rotation_slots


class HanoiRotator(RotatorBase):
    """Tower of Hanoi implementation.
    A backup slot is a power of two (1, 2, 4, 8, ...).
    """

    def __init__(self, num_rotation_slots, verbose=False):
        self.num_rotation_slots = num_rotation_slots
        self.max_rotation_slot = 2 ** (num_rotation_slots - 1)

        if verbose:
            slots = [2 ** x for x in range(num_rotation_slots)]
            logging.info("Rotation slots are: %s" % slots)

    def id_to_slot(self, rotation_id):
        adjusted_rotation_id = rotation_id % self.max_rotation_slot
        if adjusted_rotation_id == 0:
            return self.max_rotation_slot

        # find the maximum power of two that divides cleanly into
        # (rotation_id % max_rotation_slot)
        max_divisor = 1
        for divisor in (2 ** n for n in range(1, self.num_rotation_slots + 1)):
            if adjusted_rotation_id % divisor == 0:
                max_divisor = divisor
            else:
                break
        return max_divisor


class TieredRotator(RotatorBase):
    """Tiered rotation schedule.

    This is a generalization of the grandfather-father-son rotation algorithm.
    """

    def __init__(self, tier_sizes, verbose=False):
        self.num_tiers = len(tier_sizes)
        self.biggest_tier = self.num_tiers - 1
        self.tiers = []
        self.multipliers = []

        # given a list of tier sizes, configure a list of multipliers,
        # and a list of tiers with slots
        multiplier = 1
        for tier_size in tier_sizes:
            slots = [(multiplier * n) - 1 for n in range(1, tier_size + 1)]
            self.tiers.append(slots)
            self.multipliers.append(multiplier)
            multiplier *= (1 + tier_size)
        if verbose:
            logging.info("Rotation slots are: %s" % self.tiers)
            logging.info("Multipliers are: %s" % self.multipliers)

    def id_to_slot(self, rotation_id):
        slot = 0
        # try each tier, biggest multiple to smallest
        for i in reversed(range(self.num_tiers)):
            # divide the rotation_id by the multiplier .. if it is equal to
            # the multiplier-1, then we've found our slot
            (quotient, remainder) = divmod(rotation_id, self.multipliers[i])
            if remainder == (self.multipliers[i] - 1):
                # wrap within a tier (based on how many slots in the tier) to
                # prevent unlimited growth the top tier is a special case -
                # other tiers skip every nth run to promote to a higher tier
                modifier = len(self.tiers[i]) \
                    if (i == self.biggest_tier) \
                    else len(self.tiers[i]) + 1
                modified_quotient = quotient % modifier
                slot = modified_quotient \
                    * self.multipliers[i] \
                    + (self.multipliers[i] - 1)
                break
        return slot



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
archive_rotator.cli
-------------------

Handles command line arguments.

Example::

    archive-rotator -v -n 5 /path/to/foo/mydump.tgz

"""

import argparse
import logging
import os
import sys

from archive_rotator import rotator
from archive_rotator.algorithms import \
    SimpleRotator, HanoiRotator, TieredRotator


def main():
    parser = argparse.ArgumentParser(
        description='Move a file into a rotation of backup archives.')
    parser.add_argument('path', help='Path of input file to rotate')
    parser.add_argument('-n', '--num', dest='num_rotation_slots', type=int,
                        action='append',
                        help='Max number of files in the rotation')
    parser.add_argument('-v', '--verbose', dest='verbose', action="store_true",
                        help='Print info messages to stdout')
    parser.add_argument('--ext', dest='ext', default='',
                        help='Look for and preserve the named file extension')
    parser.add_argument('-d', '--destination-dir', dest='destination_dir',
                        help='Put the rotated archive in this directory.'
                             'Use if the rotated archives live in a '
                             'different directory from the source file.')
    parser.add_argument('--ignore-missing', dest='ignore_missing',
                        action="store_true",
                        help='If the input file is missing, log and exit '
                             'normally rather than exiting with an error')
    parser.add_argument('--simple', dest='simple_rotator', action="store_true",
                        help='Use the first-in-first-out rotation pattern '
                             '(default)')
    parser.add_argument('--hanoi', dest='hanoi_rotator', action="store_true",
                        help='Use the Tower of Hanoi rotation pattern')
    parser.add_argument('--tiered', dest='tiered_rotator', action="store_true",
                        help='Use the tiered rotation pattern')
    args = parser.parse_args()

    # logging
    log_level = logging.INFO if args.verbose else logging.WARN
    logging.basicConfig(stream=sys.stdout, level=log_level,
                        format='%(message)s')

    # validate args
    if not args.num_rotation_slots:
        raise ValueError("Requires at least one rotation slot.")
    for n in args.num_rotation_slots:
        if n < 1:
            raise ValueError("Values less than one are not allowed for -n.")
    if len(args.num_rotation_slots) > 1 and not args.tiered_rotator:
        raise ValueError(
            "Multiple -n values not allowed with the configured rotator.")
    if os.path.isdir(args.path):
        raise ValueError("The specified path (%s) is a directory, but must be "
                         "a file." % args.path)

    # validate path
    if not os.path.isfile(args.path):
        msg = "Specified file (%s) not found; exiting." % args.path
        if args.ignore_missing:
            logging.info(msg)
            sys.exit()
        else:
            raise IOError(msg)

    # validate ext
    if args.ext and args.ext[0] != ".":
        raise ValueError(
            "File extension (--ext) must start with the . character.")
    if args.ext and not args.path.endswith(args.ext):
        raise ValueError("The file %s does not have the file extension %s"
                         % (args.path, args.ext))

    # validate destination_dir
    if args.destination_dir and not os.path.isdir(args.destination_dir):
        raise ValueError("The specified destination directory "
                         "(--destination-dir) is not found or is not a "
                         "directory.")

    # build a rotator
    if args.hanoi_rotator:
        logging.info("Using Hanoi Rotator")
        algorithm = HanoiRotator(args.num_rotation_slots[0], args.verbose)
    elif args.tiered_rotator:
        logging.info("Using Tiered Rotator")
        algorithm = TieredRotator(args.num_rotation_slots, args.verbose)
    else:
        logging.info("Using Simple (FIFO) Rotator")
        algorithm = SimpleRotator(args.num_rotation_slots[0], args.verbose)

    # go!
    paths = rotator.Paths(args.path, args.ext, args.destination_dir)
    rotator.rotate(algorithm, paths, args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
archive_rotator.rotator
-----------------------

File moving logic goes here.
"""

from datetime import datetime
from glob import iglob
import logging
import re
import os
import shutil

FILE_NAME_GLOB = '*.backup-*'
FILE_NAME_REGEX = r'backup-(?P<rotation_id>\d+)'
FILE_NAME_TMPL = ".%(datetime_str)s.backup-%(rotation_id)d%(ext)s"
DATETIME_FORMAT = '%Y-%m-%d-%H%M%S'


class Paths(object):
    """
    Holds configuration of paths. Computes the output path.
    """

    @property
    def full_input_path(self):
        return os.path.join(self.input_dir, self.input_fn + self.input_ext)

    @property
    def output_path_no_ext(self):
        return os.path.join(self.output_dir, self.input_fn)

    @staticmethod
    def _get_now():
        # XXX: this is here to facilitate mocking in unit tests
        return datetime.now()

    def full_output_path(self, next_rotation_id):
        suffix = \
            FILE_NAME_TMPL % \
            {'datetime_str': self._get_now().strftime(DATETIME_FORMAT),
             'rotation_id': next_rotation_id,
             'ext': self.input_ext}
        return os.path.join(self.output_dir, self.input_fn + suffix)

    def __init__(self, path, ext, destination_dir):
        # if a file extension was specified, trim it off the path
        # we do not try to guess extensions, only use if explicitly configured
        path_without_ext = path if not ext else path.rsplit(ext, 1)[0]

        dir_name, file_name = os.path.split(path_without_ext)

        self.input_dir = dir_name
        self.input_fn = file_name
        self.input_ext = ext if ext else ""
        self.output_dir = destination_dir if destination_dir else dir_name


def _rotated_files(path):
    """Generator. Yields the next rotated file as a tuple:
    (path, rotation_id)
    """
    for globbed_path in iglob(path + FILE_NAME_GLOB):
        match = re.search(FILE_NAME_REGEX, globbed_path)
        if match:
            yield globbed_path, int(match.group('rotation_id'))


def _next_rotation_id(rotated_files):
    """Given the hanoi_rotator generated files in the output directory,
    returns the rotation_id that will be given to the current file. If there
    are no existing rotated files, return 0.
    """
    if not rotated_files:
        return 0
    else:
        highest_rotated_file = max(rotated_files, key=lambda x: x[1])
        return highest_rotated_file[1] + 1


def _locate_files_to_delete(algorithm, rotated_files, next_rotation_id):
    """Looks for hanoi_rotator generated files that occupy the same slot
    that will be given to rotation_id.
    """
    rotation_slot = algorithm.id_to_slot(next_rotation_id)
    for a_path, a_rotation_id in rotated_files:
        if rotation_slot == algorithm.id_to_slot(a_rotation_id):
            yield a_path


def rotate(algorithm, paths, verbose):
    # reconstruct rotation state from the file names in output dir
    rotated_files = list(_rotated_files(paths.output_path_no_ext))
    next_rotation_id = _next_rotation_id(rotated_files)
    to_delete = list(_locate_files_to_delete(algorithm,
                                             rotated_files,
                                             next_rotation_id))
    if verbose:
        rotation_slot = algorithm.id_to_slot(next_rotation_id)
        logging.info("New file: rotation_id=%s, rotation_slot=%s"
                     % (next_rotation_id, rotation_slot))

    # rename/move the input file
    from_path = paths.full_input_path
    to_path = paths.full_output_path(next_rotation_id)
    shutil.move(from_path, to_path)
    logging.info("Moved %s to %s" % (from_path, to_path))

    # remove old archives
    for f in to_delete:
        os.remove(f)
        logging.info("Removed %s" % f)

    # log the set of rotated files
    if verbose:
        logging.info("Current rotation set:")
        for f, d in _rotated_files(paths.output_path_no_ext):
            a_rotation_slot = algorithm.id_to_slot(d)
            logging.info("- %s / slot id %s" % (f, a_rotation_slot))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_integration
----------------

Integration tests. These do not cover the CLI.
"""

import contextlib
import os
import shutil
import tempfile
import unittest

from archive_rotator import rotator
from archive_rotator.algorithms import SimpleRotator, HanoiRotator


ARCHIVE_NAME = "dump.tar.gz"
EXT = ".tar.gz"


@contextlib.contextmanager
def make_temp_directory():
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)


def num_archives(temp_dir):
    return len([name for name in os.listdir(temp_dir)
                if os.path.isfile(os.path.join(temp_dir, name))])


def create_test_archive(temp_dir):
    path = os.path.join(temp_dir, ARCHIVE_NAME)
    with open(path, 'a'):
        os.utime(path, None)
    return path


class IntegrationTest(unittest.TestCase):

    def test_integration_many_iterations(self):
        with make_temp_directory() as temp_dir:
            self.assertEqual(num_archives(temp_dir), 0)
            for i in range(3):
                paths = rotator.Paths(create_test_archive(temp_dir), "", None)
                rotator.rotate(SimpleRotator(5), paths, False)
            self.assertEqual(num_archives(temp_dir), 3)
            for i in range(100):
                paths = rotator.Paths(create_test_archive(temp_dir), "", None)
                rotator.rotate(SimpleRotator(5), paths, False)
            self.assertEqual(num_archives(temp_dir), 5)

    def test_integration_with_ext(self):
        with make_temp_directory() as temp_dir:
            self.assertEqual(num_archives(temp_dir), 0)
            for i in range(3):
                paths = rotator.Paths(create_test_archive(temp_dir), EXT, None)
                rotator.rotate(SimpleRotator(20), paths, False)
            self.assertEqual(num_archives(temp_dir), 3)
            for i in range(100):
                paths = rotator.Paths(create_test_archive(temp_dir), EXT, None)
                rotator.rotate(SimpleRotator(20), paths, False)
            self.assertEqual(num_archives(temp_dir), 20)

    def test_integration_with_hanoi(self):
        with make_temp_directory() as temp_dir:
            self.assertEqual(num_archives(temp_dir), 0)
            for i in range(3):
                paths = rotator.Paths(create_test_archive(temp_dir), EXT, None)
                rotator.rotate(HanoiRotator(20), paths, False)
            self.assertEqual(num_archives(temp_dir), 3)
            for i in range(100):
                paths = rotator.Paths(create_test_archive(temp_dir), EXT, None)
                rotator.rotate(HanoiRotator(20), paths, False)
            # 1, 2, 4, 8, 16, 32, 64, 128 (first archive gets 128)
            self.assertEqual(num_archives(temp_dir), 8)

    def test_integration_with_destination_dir(self):
        with make_temp_directory() as temp_dir, \
                make_temp_directory() as temp_output_dir:
            self.assertEqual(num_archives(temp_dir), 0)
            self.assertEqual(num_archives(temp_output_dir), 0)
            for i in range(3):
                paths = rotator.Paths(create_test_archive(temp_dir), EXT,
                                      temp_output_dir)
                rotator.rotate(SimpleRotator(5), paths, False)
            self.assertEqual(num_archives(temp_dir), 0)
            self.assertEqual(num_archives(temp_output_dir), 3)
            for i in range(100):
                paths = rotator.Paths(create_test_archive(temp_dir), EXT,
                                      temp_output_dir)
                rotator.rotate(SimpleRotator(5), paths, False)
            self.assertEqual(num_archives(temp_dir), 0)
            self.assertEqual(num_archives(temp_output_dir), 5)

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_rotator
------------

Tests for `archive_rotator.rotator` module.
"""

from datetime import datetime
import unittest

from archive_rotator import rotator
from archive_rotator.algorithms import SimpleRotator


def mocked_get_now():
    """
    A method to call instead of datetime.now() to ensure testing stability
    """
    return datetime(2012, 1, 1, 10, 10, 10)


class TestPaths(unittest.TestCase):

    def test_path_parsing(self):
        paths = rotator.Paths("/a/b/~c/a.b.c", "", None)
        self.assertEqual(paths.input_dir, "/a/b/~c")
        self.assertEqual(paths.input_fn, "a.b.c")
        self.assertEqual(paths.input_ext, "")
        self.assertEqual(paths.output_dir, "/a/b/~c")

        paths = rotator.Paths("./a/b/~c/a.b.c", ".c", "/d/e/f/")
        self.assertEqual(paths.input_dir, "./a/b/~c")
        self.assertEqual(paths.input_fn, "a.b")
        self.assertEqual(paths.input_ext, ".c")
        self.assertEqual(paths.output_dir, "/d/e/f/")

    def test_property_fields(self):
        paths = rotator.Paths("/a/b/~c/a.b.c", "", None)
        self.assertEqual(paths.full_input_path, "/a/b/~c/a.b.c")
        self.assertEqual(paths.output_path_no_ext, "/a/b/~c/a.b.c")

        paths = rotator.Paths("/a/b/~c/a.b.c", ".b.c", "/d/e/f/")
        self.assertEqual(paths.full_input_path, "/a/b/~c/a.b.c")
        self.assertEqual(paths.output_path_no_ext, "/d/e/f/a")

    def test_build_output_path1(self):
        """
        no ext, no output dir, easy path
        """
        paths = rotator.Paths("foo/a.zip", "", None)
        paths._get_now = mocked_get_now

        self.assertEqual(paths.full_output_path(0),
                         "foo/a.zip.2012-01-01-101010.backup-0")

        self.assertEqual(paths.full_output_path(128),
                         "foo/a.zip.2012-01-01-101010.backup-128")

    def test_build_output_path2(self):
        """
        no ext, no output dir, harder path
        """
        paths = rotator.Paths("/a/b/~c/a.b.c", "", None)
        paths._get_now = mocked_get_now
        self.assertEqual(paths.full_output_path(0),
                         "/a/b/~c/a.b.c.2012-01-01-101010.backup-0")

    def test_build_output_path3(self):
        """
        ext, no output dir, harder path
        """
        paths = rotator.Paths("/a/b/~c/a.b.c", ".b.c", None)
        paths._get_now = mocked_get_now
        self.assertEqual(paths.full_output_path(0),
                         "/a/b/~c/a.2012-01-01-101010.backup-0.b.c")

    def test_build_output_path4(self):
        """
        no ext, output dir, harder path
        """
        paths = rotator.Paths("/a/b/~c/a.b.c", "", "/foo/bar")
        paths._get_now = mocked_get_now
        self.assertEqual(paths.full_output_path(0),
                         "/foo/bar/a.b.c.2012-01-01-101010.backup-0")


class TestRotator(unittest.TestCase):

    def test_finds_rotated_files(self):
        rotated_files = list(rotator._rotated_files("tests/test_dir_a/mydump"))
        self.assertEqual(len(rotated_files), 5)

        first_path, first_rotation_id = rotated_files[0]
        # no guaranteed order to the globbing, so this cannot verify exactly
        self.assertRegexpMatches(
            first_path,
            "tests\/test_dir_a\/mydump.*2015-12-02.*\.backup-[3-7].*")
        self.assertGreaterEqual(first_rotation_id, 3)

    def test_rotated_files_respects_name(self):
        # there are no files matching "another_name"
        rotated_files = list(rotator._rotated_files("tests/test_dir_a/foo"))
        self.assertEqual(len(rotated_files), 0)

    def test_finds_next_rotation_id(self):
        rotated_files = list(rotator._rotated_files("tests/test_dir_a/mydump"))
        x = rotator._next_rotation_id(rotated_files)
        self.assertEqual(x, 8)

    def test_next_rotation_id_is_0_without_match(self):
        rotated_files = list(rotator._rotated_files("tests/test_dir_a/foo"))
        x = rotator._next_rotation_id(rotated_files)
        self.assertEqual(x, 0)

    def test_locate_files_to_delete(self):
        rotated_files = list(rotator._rotated_files("tests/test_dir_a/mydump"))

        # 5 slots
        to_delete = list(rotator._locate_files_to_delete(
            SimpleRotator(5), rotated_files, 8))
        self.assertEqual(len(to_delete), 1)

        # 10 slots (don't delete anything)
        to_delete = list(rotator._locate_files_to_delete(
            SimpleRotator(10), rotated_files, 8))
        self.assertEqual(len(to_delete), 0)

        # 2 slots (delete a #4 and #6)
        to_delete = list(rotator._locate_files_to_delete(
            SimpleRotator(2), rotated_files, 8))
        self.assertEqual(len(to_delete), 2)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]

# https://pypi.python.org/pypi?%3Aaction=list_classifiers
setup(
    name='archive-rotator',
    version='0.2.0',
    description="Flexible utility for rotating backup files.",
    long_description=readme + '\n\n' + history,
    author="Max Harper",
    author_email='maxharp3r@gmail.com',
    url='https://github.com/maxharp3r/archive-rotator',
    packages=[
        'archive_rotator',
    ],
    package_dir={'archive_rotator': 'archive_rotator'},
    entry_points={
        'console_scripts': [
            'archive-rotator = archive_rotator.cli:main',
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT License",
    zip_safe=False,
    keywords='backup rotation',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: System :: Archiving :: Backup',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_algorithms
---------------

Tests for `archive_rotator.algorithms` module.
"""

import unittest

from archive_rotator.algorithms import \
    SimpleRotator, HanoiRotator, TieredRotator


class TestAlgorithms(unittest.TestCase):

    def test_simple_rotator_assignment(self):
        self.assertEqual(SimpleRotator(4).id_to_slot(0), 0)
        self.assertEqual(SimpleRotator(4).id_to_slot(7), 3)

    def test_hanoi_rotator_assignment(self):
        # The first backup (id 0) should be assigned the biggest slot
        # available: 2^(n-1)
        self.assertEqual(HanoiRotator(3).id_to_slot(0), 4)
        self.assertEqual(HanoiRotator(4).id_to_slot(0), 8)
        self.assertEqual(HanoiRotator(5).id_to_slot(0), 16)

        self.assertEqual(HanoiRotator(4).id_to_slot(1), 1)
        self.assertEqual(HanoiRotator(4).id_to_slot(3), 1)
        self.assertEqual(HanoiRotator(4).id_to_slot(9), 1)
        self.assertEqual(HanoiRotator(4).id_to_slot(2), 2)
        self.assertEqual(HanoiRotator(4).id_to_slot(18), 2)
        self.assertEqual(HanoiRotator(4).id_to_slot(4), 4)
        self.assertEqual(HanoiRotator(4).id_to_slot(8), 8)
        self.assertEqual(HanoiRotator(4).id_to_slot(24), 8)

        self.assertEqual(HanoiRotator(5).id_to_slot(16), 16)
        self.assertEqual(HanoiRotator(5).id_to_slot(30), 2)
        self.assertEqual(HanoiRotator(5).id_to_slot(32), 16)

        self.assertEqual(HanoiRotator(6).id_to_slot(32), 32)

    def test_tiered_rotator_assignment(self):
        self.assertEqual(TieredRotator([2, 2]).id_to_slot(0), 0)
        self.assertEqual(TieredRotator([2, 2]).id_to_slot(1), 1)
        self.assertEqual(TieredRotator([2, 2]).id_to_slot(2), 2)
        self.assertEqual(TieredRotator([2, 2]).id_to_slot(8), 2)
        self.assertEqual(TieredRotator([2, 2]).id_to_slot(9), 0)

        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(0), 0)
        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(1), 1)
        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(4), 4)
        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(5), 0)
        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(19), 19)
        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(39), 39)

        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(10), 0)
        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(24), 4)
        self.assertEqual(TieredRotator([4, 3, 2]).id_to_slot(59), 19)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())
